#include "My_mvt.h"
#include "VT.h"
#include "Burden.h"
#include "Constant.h"
#include "math.h"
#include <string>
#include <cstring>
#include <iostream>
#include <stdlib.h>
#include <algorithm>
#include <time.h>
#include "Random.h"
#define MATHLIB_STANDALONE 
#include <Rmath.h>

bool VT::asymptotic = false;
bool VT::permuteBool = false;
bool VT::vcfAnnotated = true;
bool VT::preProcess = false;
String VT::chopTo = "chop_to";
int VT::chop = 20;
int VT::permuteMin = 1000;
int VT::permuteMax = 3000000;
bool VT::VTwtMB = false;
bool VT::VTwtCMC_binary = false;
bool VT::VTwtCMC_counts = true;
bool VT::noStep = false;


VT::VT(){}

VT::~VT()
{}

bool VT::GetTmpGenoMatrixPED(FastTransform & trans, Pedigree & ped,GroupFromAnnotation & group,int groupNum)
{
   bool status = false;

   mac =0.0;
   numVarPerGene=0;

   tmpGenoMatrix.resize(0,0);
   varsIncluded.Clear();
   mafIncluded.Clear();
   Vector mafSqrIncluded;

   for(int m=0;m<group.chrList[groupNum].Length();m++)
   {
      String markerName = group.chrList[groupNum][m];
      markerName += ":";
      markerName += group.posList[groupNum][m];
      int markerid = ped.markerLookup.Integer(markerName);
      if(markerid==-1)
      {
	 continue;
      }
      int alleleCount = CountAlleles(ped, markerid);
      if(alleleCount!=2)
	 continue;

      String refAllele = ped.GetMarkerInfo(markerid)->alleleLabels[1];
      String altAllele = ped.GetMarkerInfo(markerid)->alleleLabels[2];

      String allele = refAllele + "/" + altAllele;
      if(allele != group.alleleList[groupNum][m])
      {
	 String flipAllele = altAllele + "/" + refAllele;
	 if(flipAllele != group.alleleList[groupNum][m])
	 {
	    continue;
	 }
      }
      //if marker m from group list is found then get the genotype vector
      bool XStatus = false;
      if(AutoFit::fitX)
      {
	 int found = markerName.FastFind(SVT::xLabel,0);
	 if(found!=-1)
	 {
	    if(group.posList[groupNum][m] > SVT::Xstart && group.posList[groupNum][m] < SVT::Xend)
	    {
	       XStatus=true;
	    }
	 }
      }

      int n=0,nf=0;
      double fmaf = 0.0, maf=0.0;
      Vector genotype;
      genotype.Dimension(trans.persons);
      for(int i=0;i<trans.persons;i++)
	 genotype[i]=-1.0;

      //Loop through samples according to trans.pPheno 
      int idx=0;
double missing=0.0;
      for(int f=0;f<ped.familyCount;f++)
      {
	 for(int p=0;p<trans.pPheno[f].Length();p++)
	 {
	    if(!ped[trans.pPheno[f][p]].isGenotyped(markerid))
	    {
	       idx++;
missing++;
	       continue;
	    }
	    if(ped[trans.pPheno[f][p]].markers[markerid].isHeterozygous())
	    {
	       if(XStatus)
	       {
		  if(ped[trans.pPheno[f][p]].sex==SVT::femaleLabel)
		  {
		     genotype[idx] = 1.0;
		  }
		  if(ped[trans.pPheno[f][p]].sex==SVT::maleLabel)
		  {
		     genotype[idx] = 2.0;
		  }
	       }
	       else
	       {
		  genotype[idx] = 1.0;
	       }
	    }
	    else if(ped[trans.pPheno[f][p]].markers[markerid].isHomozygous())
	    {
	       if(ped[trans.pPheno[f][p]].markers[markerid].hasAllele(2))
	       {
		  genotype[idx] = 2.0;
	       }
	       else if(ped[trans.pPheno[f][p]].markers[markerid].hasAllele(1))
	       {
		  genotype[idx] = 0.0;
	       }
	    }
	    maf+=genotype[idx];
	    n++;
	    int founder = trans.foundersHash.Integer(ped[trans.pPheno[f][p]].pid);
	    if(founder != -1)
	    {
	       fmaf += genotype[idx];
	       nf++;
	    }
	    idx++;
	 }
      }
      maf /= 2.0*n;
      fmaf /= 2.0*nf;

      if(fmaf==0.0 && maf>0)
	 fmaf = 1.0/(2.0*nf);
missing /= trans.persons;

if(fmaf==0.0 || fmaf==1.0 || maf==0.0 || maf==1.0)
continue;

      //check if maf falls into the range
      if(!group.freqOut)
      {
	 if(SVT::FounderFreq)
	 {
	    double fmaf_tmp;
	    fmaf_tmp = fmaf;
	    if(fmaf>0.5)
	       fmaf_tmp = 1.0-fmaf;
	    if(fmaf_tmp>Burden::mafMax || fmaf_tmp<Burden::mafMin)
	       continue;
	 }
	 else
	 {
	    double maf_tmp;
	    maf_tmp=maf;
	    if(maf>0.5)
	       maf_tmp = 1.0-maf;
	    if(maf_tmp>Burden::mafMax || maf_tmp<Burden::mafMin)
	       continue;
	 }
      }
if(missing>Burden::miss)
continue;

      if(maf>0.5 && maf>1.0-Burden::mafMax)
      {
         status=true;
      }
      else if(maf<0.5 && maf<Burden::mafMax)
      {
         status = true;
      }
      else
      {
         continue;
      }
      numVarPerGene++;

      bool flip=false;
      if(SVT::FounderFreq && fmaf>0.5)
      {
	 flip=true;
      }
      else if(maf>0.5)
      {
	 flip=true;
      }
      if(flip)
      {
	 fmaf = 1.0-fmaf;
	 maf = 1.0-maf;
	 for(int i=0;i<trans.persons;i++)
	 {
	    if(genotype[i]!=-1.0)
	    {
	       genotype[i] = 2.0-genotype[i];
	    }
	 }
      }

      for(int i=0;i<trans.persons;i++)
      {
	 if(genotype[i]!=-1.0)
	 {
	    mac+=genotype[i];
	 }
      }

      if(Burden::fullResult)
      {
	 String tmp = "";
	 tmp += group.chrList[groupNum][m];
	 tmp+=":";
	 tmp+=group.posList[groupNum][m];
	 tmp+= "_";
	 if(flip)
	 {
	    tmp+=altAllele;
	    tmp+="/";
	    tmp += refAllele;
	 }
	 else
	 {
	    tmp+=refAllele;
	    tmp+="/";
	    tmp += altAllele;
	 }
	 varsIncluded.Push(tmp);
      }

      if(SVT::FounderFreq)
      {
	 mafIncluded.Push(fmaf);
	 mafSqrIncluded.Push(fmaf*fmaf);
      }
      else
      {
	 mafIncluded.Push(maf);
	 mafSqrIncluded.Push(maf*maf);
      }

      tmpGenoMatrix.conservativeResize(trans.persons,numVarPerGene);

      //update tmpGenoMatrix
      for(int row=0;row<trans.persons;row++)
      {
	 tmpGenoMatrix(row,numVarPerGene-1) = genotype[row];
      }
   }
   if(numVarPerGene==0)
   {
      status =false;
      return status;
   }
   if(mac<Burden::MAC)
   {
      status = false;
      return status;
   }

   return status;
}

bool VT::GetTmpGenoMatrixDosage(FastTransform & trans, Pedigree & ped,GroupFromAnnotation & group,int groupNum)
{
   bool status = false;

   mac =0.0;
   numVarPerGene=0;

   tmpGenoMatrix.resize(0,0);
   varsIncluded.Clear();
   mafIncluded.Clear();
   Vector mafSqrIncluded;

   for(int c=0;c<group.uniqueChr[groupNum].Length();c++)
   {
      reader.set1BasedReadSection(group.uniqueChr[groupNum][c].c_str(),group.startPos[groupNum][c],group.endPos[groupNum][c]+1);

   while(reader.readRecord(record))
   {
      int m;
      int position = record.get1BasedPosition();
      String altAllele = record.getAltStr();
      String refAllele = record.getRefStr();
      bool match = false;
      for(int j=0;j<group.chrList[groupNum].Length();j++)
      {
	 if(position == group.posList[groupNum][j])
	 {
	    if(group.uniqueChr[groupNum][c] == group.chrList[groupNum][j])
	    {
	       match=true;
	       m = j;
	       break;
	    }
	 }
      }
      if(!match)
	 continue;

      String allele = refAllele + "/" + altAllele;
      if(allele != group.alleleList[groupNum][m])
      {
	 String flipAllele = altAllele + "/" + refAllele;
	 if(flipAllele != group.alleleList[groupNum][m])
	    continue;
	 else
	    printf("ref and alt alleles are flipped!");
      }

      int n=0,nf=0;
      double fmaf = 0.0, maf=0.0;
      VcfRecordGenotype & genoInfo = record.getGenotypeInfo();

      bool XStatus = false;
      if(AutoFit::fitX)
      {
	 String chr = group.chrList[groupNum][0].c_str();
	 if(chr==SVT::xLabel && position > SVT::Xstart && position < SVT::Xend)
	 {
	    XStatus=true;
	 }
      }

      Vector genotype;
      genotype.Dimension(trans.persons);
      for(int i=0;i<trans.persons;i++)
	 genotype[i]=-1.0;

      //Loop through samples according to trans.pPheno 
      //read in dosage into genotype vector
      int idx=0;
      double missing = 0.0;
      for (int f = 0; f < ped.familyCount; f++)
      {
	 for(int j=0;j<trans.pPheno[f].Length();j++)
	 {
	    String sample = ped[trans.pPheno[f][j]].pid;

	    //get the sample number by sample ID from hash
	    int s = trans.sampleVCFIDHash.Integer(sample);
	    if(s!=-1)
	    {
	       const std::string * geno = genoInfo.getString("DS",s);
	       //if dosage is not missing
	       if(strcmp((*geno).c_str(),".")>0)
	       {
		  float dose = atof((*geno).c_str());
		  if(dose >=0)
		  {
		     if(XStatus && dose >0 && ped[trans.pPheno[f][j]].sex==SVT::maleLabel)
			genotype[idx] = 2;
		     else
			genotype[idx] = dose;

		     maf += genotype[idx];
		     n++;

		     //judge if this is a founder
		     int founder = trans.foundersHash.Integer(sample);
		     if(founder != -1)
		     {
			fmaf += genotype[idx];
			nf++;
		     }
		  }
	       }
	       else
		  missing++;
	    }
	    else
	    {
	       error("FATAL ERROR! Sample %s is not included in VCF file.\n",sample.c_str());
	    }
	    idx++;
	 }
      }
      maf /= 2.0*n;
      fmaf /= 2.0*nf;

      if(fmaf==0.0 && maf>0)
	 fmaf = 1.0/(2.0*nf);

      missing /= trans.persons;
      if(missing>Burden::miss)
	 continue;
      if(fmaf==0.0 || fmaf==1.0 || maf==0.0 || maf==1.0)
	 continue;

      //check if maf falls into the range
      if(!group.freqOut)
      {
	 if(SVT::FounderFreq)
	 {
	    double fmaf_tmp;
	    fmaf_tmp = fmaf;
	    if(fmaf>0.5)
	       fmaf_tmp = 1.0-fmaf;
	    if(fmaf_tmp> Burden::mafMax || fmaf_tmp<Burden::mafMin)
	       continue;
	 }
	 else
	 {
	    double maf_tmp;
	    maf_tmp=maf;
	    if(maf>0.5)
	       maf_tmp = 1.0-maf;
	    if(maf_tmp>Burden::mafMax || maf_tmp<Burden::mafMin)
	       continue;
	 }
      }

      if(maf>0.5 && maf>1.0-Burden::mafMax)
      {
	 status=true;
      }
      else if(maf<0.5 && maf<Burden::mafMax)
      {
	 status = true;
      }
      else
      {
	 continue;
      }
      numVarPerGene++;

      bool flip=false;
      if(SVT::FounderFreq && fmaf>0.5)
      {
	 flip=true;
      }
      else if(maf>0.5)
      {
	 flip=true;
      }
      if(flip)
      {
	 fmaf = 1.0-fmaf;
	 maf = 1.0-maf;
	 for(int i=0;i<trans.persons;i++)
	 {
	    if(genotype[i]!=-1.0)
	    {
	       genotype[i] = 2.0-genotype[i];
	    }
	 }
      }

      for(int i=0;i<trans.persons;i++)
      {
	 if(genotype[i]!=-1.0)
	 {  
	    mac+=genotype[i];
	 }  
      }

      if(Burden::fullResult)
      {
	 String tmp = "";
	 tmp += group.chrList[groupNum][m];
	 tmp+=":";
	 tmp+=group.posList[groupNum][m];
	 tmp+= "_";
	 if(flip)
	 {
	    tmp+=altAllele;
	    tmp+="/";
	    tmp += refAllele;
	 }
	 else
	 {
	    tmp+=refAllele;
	    tmp+="/";
	    tmp += altAllele;
	 }
	 varsIncluded.Push(tmp);
      }

      if(SVT::FounderFreq)
      {
	 mafIncluded.Push(fmaf);
	 mafSqrIncluded.Push(fmaf*fmaf);
      }
      else
      {
	 mafIncluded.Push(maf);
	 mafSqrIncluded.Push(maf*maf);
      }

      tmpGenoMatrix.conservativeResize(trans.persons,numVarPerGene);

      //update tmpGenoMatrix
      for(int row=0;row<trans.persons;row++)
      {
	 tmpGenoMatrix(row,numVarPerGene-1) = genotype[row];
      }
   }
   }

   if(numVarPerGene==0)
   {
      status =false;
      return status;
   }
   if(mac<Burden::MAC)
   {
      status = false;
      return status;
   }

   return status;
}

bool VT::GetTmpGenoMatrixVCF(FastTransform & trans, Pedigree & ped,GroupFromAnnotation & group,int groupNum)
{
   bool status = false;

   varsIncluded.Clear();
   //mac is the total number of alternative alleles included
   mac = 0.0;
   //numVarPerGene is the total number of variants included in this gene.
   numVarPerGene=0;

   mafIncluded.Clear();
   Vector mafSqrIncluded;

   for(int c=0;c<group.uniqueChr[groupNum].Length();c++)
   {
      reader.set1BasedReadSection(group.uniqueChr[groupNum][c].c_str(),group.startPos[groupNum][c],group.endPos[groupNum][c]+1);

      while(reader.readRecord(record))
      {
         int m;
         int position = record.get1BasedPosition();
         String altAllele = record.getAltStr();
         String refAllele = record.getRefStr();
         bool match = false;
         for(int j=0;j<group.chrList[groupNum].Length();j++)
         {
            if(position == group.posList[groupNum][j])
            {
               if(group.uniqueChr[groupNum][c] == group.chrList[groupNum][j])
               {
                  match=true;
                  m = j;
                  break;
               }
            }
         }
         if(!match)
            continue;

	 bool XStatus = false;

	 if(AutoFit::fitX)
	 {
	    String chr = group.chrList[groupNum][0].c_str();
	    if(chr==SVT::xLabel && position > SVT::Xstart && position < SVT::Xend)
	    {
	       XStatus=true;
	    }
	 }
	    String allele = refAllele + "/" + altAllele;
	    if(allele != group.alleleList[groupNum][m])
	    {
	       String flipAllele = altAllele + "/" + refAllele;
	       if(flipAllele != group.alleleList[groupNum][m])
		  continue;
	       else
	       {
		  printf("ref and alt alleles are flipped!");
	       }
	    }

	 double fmaf=0.0,maf=0.0;
	 int n=0,nf=0;

	 //Fill in genotype vector
	 Vector genotype;
	 genotype.Dimension(trans.persons);
	 for(int i=0;i<trans.persons;i++)
	    genotype[i] = -1.0;

	 int idx=0;
	 double missing = 0.0;
	 for (int f = 0; f < ped.familyCount; f++)
	 {
	    for(int p=0;p<trans.pPheno[f].Length();p++)
	    {
	       String sample = ped[trans.pPheno[f][p]].pid.c_str();

	       //get the sample number by sample ID from hash
	       int s = trans.sampleVCFIDHash.Integer(sample);
	       if(s!=-1)
	       {
		  int numGTs = record.getNumGTs(s);
		  Vector allele;

		  for(int j = 0; j < numGTs; j++)
		  {
		     allele.Push(record.getGT(s,j));
		  }
		  int bad =0;
		  for(int m=0;m<allele.Length();m++)
		  {
		     bad += (allele[m]==VcfGenotypeSample::MISSING_GT);
		  }
		  //if no missing allele exists
		  //if no missing allele exists
		  if(bad==0)
		  {
		     int sum=0;
		     for(int i=0;i<allele.Length();i++)
			sum+=allele[i];

		     if(XStatus && sum >0 && ped[trans.pPheno[f][p]].sex==SVT::maleLabel)
			genotype[idx] = 2;
		     else
			genotype[idx] = sum;

		     maf += genotype[idx];
		     n++;
		     //judge if this is a founder
		     int founder = trans.foundersHash.Integer(sample);
		     if(founder!=-1)
		     {
			nf++;
			fmaf+=genotype[idx];
		     }
		  }
		  else
		     missing++;
	       }
	       idx++;
	    }
	 }
	 //Adjust genotype here
	 maf /= 2.0*n;
	 fmaf /=2.0*nf;

	 if(fmaf==0.0 && maf>0)
	    fmaf = 1.0/(2.0*nf);

	 missing /= trans.persons;
	 if(missing>Burden::miss)
	    continue;
	 if(fmaf==0.0 || fmaf==1.0 || maf==0.0 || maf==1.0)
	    continue;

	 //check if maf falls into the range
	 if(!group.freqOut)
	 {
	    if(SVT::FounderFreq)
	    {
	       double fmaf_tmp;
	       fmaf_tmp = fmaf;
	       if(fmaf>0.5)
		  fmaf_tmp = 1.0-fmaf;
	       if(fmaf_tmp>Burden::mafMax || fmaf_tmp<Burden::mafMin)
		  continue;
	    }
	    else
	    {
	       double maf_tmp;
	       maf_tmp=maf;
	       if(maf>0.5)
		  maf_tmp = 1.0-maf;
	       if(maf_tmp>Burden::mafMax || maf_tmp<Burden::mafMin)
		  continue;
	    }
	 }

	 if(maf>0.5 && maf>1.0-Burden::mafMax)
	 {
	    status=true;
	 }
	 else if(maf<0.5 && maf<Burden::mafMax)
	 {
	    status = true;
	 }
	 else
	 {
	    continue;
	 }
	 numVarPerGene++;

	 bool flip=false;
	 if(SVT::FounderFreq && fmaf>0.5)
	 {
	    flip=true;
	 }
	 else if(maf>0.5)
	 {
	    flip=true;
	 }
	 if(flip)
	 {
	    fmaf = 1.0-fmaf;
	    maf = 1.0-maf;
	    for(int i=0;i<trans.persons;i++)
	    {
	       if(genotype[i]!=-1.0)
	       {
		  genotype[i] = 2.0-genotype[i];
	       }
	    }
	 }

	 for(int i=0;i<trans.persons;i++)
	 {
	    if(genotype[i]!=-1.0)
	    {  
	       mac+=genotype[i];
	    }  
	 }

	 if(Burden::fullResult)
	 {
	    String tmp = "";
	    tmp += group.chrList[groupNum][m];
	    tmp+=":";
	    tmp+=group.posList[groupNum][m];
	    tmp+= "_";
	    if(flip)
	    {
	       tmp+=altAllele;
	       tmp+="/";
	       tmp += refAllele;
	    }
	    else
	    {
	       tmp+=refAllele;
	       tmp+="/";
	       tmp += altAllele;
	    }
	    varsIncluded.Push(tmp);
	 }

	 if(SVT::FounderFreq)
	 {
	    mafIncluded.Push(fmaf);
	    mafSqrIncluded.Push(fmaf*fmaf);
	 }
	 else
	 {
	    mafIncluded.Push(maf);
	    mafSqrIncluded.Push(maf*maf);
	 }
	 tmpGenoMatrix.conservativeResize(trans.persons,numVarPerGene);

	 //update tmpGenoMatrix
	 for(int row=0;row<trans.persons;row++)
	 {
	    tmpGenoMatrix(row,numVarPerGene-1) = genotype[row];
	 }
      }
   }

   if(numVarPerGene==0)
   {
      status =false;
      return status;
   }
   if(mac<Burden::MAC)
   {
      status = false;
      return status;
   }

   return status;
}

bool VT::GetGenotypeMatrix(FastTransform & trans, Pedigree & ped,GroupFromAnnotation & group,int groupNum)
{
   bool status = false;
   tmpGenoMatrix.resize(0,0);

   if(SVT::genoFromPed)
      status = GetTmpGenoMatrixPED(trans,ped,group,groupNum);
   if(SVT::genoFromVCF)
      status = GetTmpGenoMatrixVCF(trans,ped,group,groupNum);
   if(SVT::dosage)
      status = GetTmpGenoMatrixDosage(trans,ped,group,groupNum);

   //adjust the genotypeMatrix by its mean
   for(int col=0;col<numVarPerGene;col++)
   {
      double mean =0.0;
      int n=0;
      for(int i=0;i<trans.persons;i++)
      {
	 if(tmpGenoMatrix(i,col)!=-1.0)
	 {
	    mean+=tmpGenoMatrix(i,col);
	    n++;
	 }
      }
      mean /=n;
      for(int i=0;i<trans.persons;i++)
      {
	 if(tmpGenoMatrix(i,col)==-1.0)
	 {
	    tmpGenoMatrix(i,col) = 0.0;
	 }
	 else
	    tmpGenoMatrix(i,col) -=mean;
      }
   }

   return status;
}

bool VT::FinalizeGenoMatrix(FastTransform & trans,bool reduceDimension)
{
   //printf("Finalizing genotype matrix.\n");
   //work on mafIncluded to get unique cutoffs
   bool status = true;
   if(mafIncluded.Length()==1)
   {
      unique_maf.Copy(mafIncluded);
      transGeno = trans.UD *tmpGenoMatrix;
      selectionIndicator.Dimension(unique_maf.Length(),mafIncluded.Length(),1);
      return status;
   }
   unique_maf.Copy(mafIncluded);
   unique_maf.RemoveDuplicates();
   unique_maf.Sort();

   if(reduceDimension)
   {
      while(unique_maf.Length()>20)
      {
	 unique_maf.Delete(0);
      }
   }
   Eigen::MatrixXf genotypeMatrix;
   genotypeMatrix.resize(trans.persons,unique_maf.Length());
   genotypeMatrix.setZero();
   selectionIndicator.Dimension(unique_maf.Length(),mafIncluded.Length(),0);

   for(int j=0;j<unique_maf.Length();j++)
   {
      for(int i=0;i<mafIncluded.Length();i++)
      {
	 if(mafIncluded[i]<=unique_maf[j])
	 {
	    selectionIndicator[j][i]=1;
	    for(int p=0;p<trans.persons;p++)
	    {
	       genotypeMatrix(p,j) += tmpGenoMatrix(p,i);
	    }
	 }
      }
   }

   /*
      if(VTwtCMC_binary)
      {
      for(int p=0;p<trans.persons;p++)
      {
      if(genotypeMatrix(p,j)>0.0)
      genotypeMatrix(p,j) = 1.0;
      }
      }
    */
   if(genotypeMatrix.cols()==0)
   {
      status = false;
      return status;
   }

   transGeno = trans.UD*genotypeMatrix;
   return status;
}

void VT::CalculateAsymptotic(Pedigree & ped, FastTransform & trans,FastFit & engine,String & traitName,GroupFromAnnotation & group,FILE * log)
{
   //double sigma_g2 = engine.sigma_g2Hat;
   //double sigma_e2 = engine.sigma_e2Hat;
   //double h2 = sigma_g2/(sigma_g2+sigma_e2);
   // printf("g,e,var:%g,%g,%g\n",sigma_g2,sigma_e2,trans.traitVar);
   double h2 = engine.sigma_g2Hat/trans.traitVar;

   FILE * VTout;

   String VTfilename;
   VTfilename = Burden::outputFile + traitName + ".VT.Association.txt";
   VTout = fopen(VTfilename,"wt");

   if(VTwtMB + VTwtCMC_binary + VTwtCMC_counts ==1)
   {
      if(VTwtMB)
      {
	 fprintf(VTout,"\nWeight used is Madsen-Browning.\n");
      } 
      else if(VTwtCMC_binary)
      {
	 fprintf(VTout,"\nWeight used is VTwtCMC_binary.\n");
      } 
      else
      {
	 fprintf(VTout,"\nWeight used is VTwtCMC_counts.\n");
      } 
   }
   else
   {
      fprintf(VTout,"\nWeights used are: %s%s%s.\n",VTwtMB?"Madsen-Browning ":"",VTwtCMC_binary?"VTwtCMC_binary ":"",VTwtCMC_counts?"VTwtCMC_counts ":"");
   }

   fprintf(VTout,"\n%s mean=%6.2f, variance=%6.2f, heritability=%6.2f \n",traitName.c_str(),trans.traitMean,trans.traitVar,h2*100.0);
   fprintf(VTout,"===================================================================\n");
   fprintf(VTout,"Gene\tNumVarIncluded\tMafCutoff\tActualThreshold\tVarsIncluded\tMAFs\tSingleVarPvalues\tSingleVarEffs\teffSize\tpvalue\n");

   int bad =0;
   Eigen::MatrixXf rh;
   rh = (trans.inv*trans.UDX.transpose()).transpose();

   for(int i=0;i<group.annoGroups.Length();i++)
   {
      if(i>100 && i%100==1)
	 printf("  Finished analyzing %d genes.\n",((int)i/100)*100);
      String geneName = group.annoGroups[i];
      //printf("Testing gene %s.\n",geneName.c_str());

      if(GetGenotypeMatrix(trans,ped,group,i))
      {
	 if(Burden::fullResult)
	    CalculateSingleVarPvalue(trans,geneName,engine,tmpGenoMatrix);

	 if(!FinalizeGenoMatrix(trans,false))
	 {
	    //print output here
	    bad++;
	    fprintf(log,"Gene %s is skipped due to no qualified markers.\n",geneName.c_str());
	    continue;
	 }

	 if(mafIncluded.Length()==1)
	 {
	    fprintf(VTout,"%s\t1\t%g\t%g\t%s\t",geneName.c_str(),Burden::mafMax,mafIncluded[0],varsIncluded[0].c_str());
	    fprintf(VTout,"%g\t%g\t%g\t%g\t%g\n",mafIncluded[0],singlevarPvalue[0],singlevarEff[0],singlevarEff[0],singlevarPvalue[0]);
	    continue;
	 }
      }
      else 
      {
	 //print output here
	 bad++;
	 fprintf(log,"Gene %s is skipped due to no qualified markers.\n",geneName.c_str());
	 continue;
      }

      //calcualte U and V
      double pvalue;
      Eigen::VectorXf score,U,effSize;
      score.resize(transGeno.cols());
      effSize.resize(transGeno.cols());
      U.resize(transGeno.cols());
      U = transGeno.transpose()*residuals_original; 

      Eigen::MatrixXf V;
      V.resize(transGeno.cols(),transGeno.cols());

      V = transGeno.transpose()*trans.projection*transGeno;
      //      V = transGeno.transpose()*transGeno;

      for(int i=0;i<transGeno.cols();i++)
      {
	 score[i] = fabs(U[i]/sqrt(V(i,i)));
	 effSize[i] = U[i]/V(i,i);
      }

      int cutoff = transGeno.cols();
      //get the chosen cutoff and output to .log file
      double t_max = score[0];
      int chosen_cutoff=0;
      for(int i=1;i<cutoff;i++)
      {
	 if(t_max < score[i])
	 {
	    t_max = score[i];
	    chosen_cutoff = i;
	 }
      }
      //double t_max = score.Max()/sqrt(sigma2_hat);

      String method;
      if(VTwtMB + VTwtCMC_binary + VTwtCMC_counts ==1)
      {
	 if(VTwtMB) 
	    method = "MB";
	 if(VTwtCMC_binary)
	    method = "VTwtCMC_binary";
	 if(VTwtCMC_counts)
	    method = "VTwtCMC_counts";
      }
      else if(VTwtMB && VTwtCMC_binary && !VTwtCMC_counts)
      {
	 if(chosen_cutoff < length)
	    method = "MB";
	 else 
	    method = "VTwtCMC_binary";
      }
      else if(VTwtMB && VTwtCMC_counts && !VTwtCMC_binary)
      {
	 if(chosen_cutoff < length)
	    method = "MB";
	 else
	    method = "VTwtCMC_counts";
      }
      else if(VTwtCMC_counts && VTwtCMC_binary && !VTwtMB)
      {
	 if(chosen_cutoff < length)
	    method = "VTwtCMC_binary";
	 else
	    method = "VTwtCMC_counts";
      }
      else if(VTwtMB + VTwtCMC_binary + VTwtCMC_counts ==3)
      {
	 if(chosen_cutoff < length)
	    method = "MB";
	 else if(chosen_cutoff < 2*length)
	    method = "VTwtCMC_binary";
	 else 
	    method = "VTwtCMC_counts";
      }

      //	    printf("chosen_cutoff is:%d\n",chosen_cutoff);
      double selected_maf = unique_maf[chosen_cutoff];

      double * lower = new double [cutoff];
      double * upper = new double [cutoff];
      double * mean = new double [cutoff];

      for(int i=0;i<cutoff;i++)
      {
	 mean[i] = 0.0;
	 lower[i] = -t_max;
	 upper[i] = t_max;
      }

      Matrix cov(transGeno.cols(),transGeno.cols());
      cov2cor(V,cov);
      /*
	 printf("t_max is: %g\n",t_max);
	 printf("V is:\n");
	 for(int i=0;i<V.rows();i++)
	 {
	 for(int j=0;j<V.cols();j++)
	 {
	 printf("%g\t",V(i,j));
	 }
	 printf("\n");
	 }
	 printf("cov is:\n");
	 for(int i=0;i<cov.rows;i++)
	 {
	 for(int j=0;j<cov.cols;j++)
	 {
	 printf("%g\t",cov[i][j]);
	 }
	 printf("\n");
	 }
       */
      //Use pmvnorm to calculate the asymptotic p-value
      Vector result;
      pmvnorm(lower,upper,mean,cov,false,result);
      //printf("pmvnorm results:%g,%g,%g\n",result[0],result[1],result[2]);

      String vars="";
      for(int id=0;id<selectionIndicator[chosen_cutoff].Length();id++)
      {
	 if(selectionIndicator[chosen_cutoff][id]==1)
	 {
	    vars += varsIncluded[id];
	    if(id<selectionIndicator[chosen_cutoff].Length()-1)
	       vars += ";";
	 }
      }

      pvalue = 1.0 - result[0];
      if(pvalue==2.0)
      {
	 fprintf(VTout,"%s\t%d\t%g\t%g\t%s\t",geneName.c_str(),(int)selectionIndicator[chosen_cutoff].Sum(),Burden::mafMax,selected_maf,vars.c_str());
	 for(int id=0;id<selectionIndicator[chosen_cutoff].Length();id++)
	 {
	    if(selectionIndicator[chosen_cutoff][id]==1)
	       fprintf(VTout,"%g;",mafIncluded[id]);
	 }
	 fprintf(VTout,"\t");
	 for(int id=0;id<selectionIndicator[chosen_cutoff].Length();id++)
	 {
	    if(selectionIndicator[chosen_cutoff][id]==1)
	       fprintf(VTout,"%g;",singlevarPvalue[id]);
	 }
	 fprintf(VTout,"\t");
	 for(int id=0;id<selectionIndicator[chosen_cutoff].Length();id++)
	 {
	    if(selectionIndicator[chosen_cutoff][id]==1)
	       fprintf(VTout,"%g;",singlevarEff[id]);
	 }
	 fprintf(VTout,"\t%g",effSize[chosen_cutoff]);
	 fprintf(VTout,"\tERROR:CORR_MATRIX_NOT_POS_SEMI_DEF\n");
	 continue;
      }

      if(pvalue==0.0)
      {
	 //printf("using Shuang's algorithm for mvnorm prob ... ");
	 //reduce dimension here
	 //double test = 0.9999975;
	 //double aftertest =1.0-test;
	 //printf("test value is %g %g\n",aftertest,test);
	 if(unique_maf.Length()>20)
	    FinalizeGenoMatrix(trans,true);
	 pvalue = CalculatePvalue(score,cov,t_max);
	 //printf("completed!\n");
      }

      fprintf(VTout,"%s\t%d\t%g\t%g\t%s\t",geneName.c_str(),(int)selectionIndicator[chosen_cutoff].Sum(),Burden::mafMax,selected_maf,vars.c_str());
      for(int id=0;id<selectionIndicator[chosen_cutoff].Length();id++)
      {
	 if(selectionIndicator[chosen_cutoff][id]==1)
	 {
	    if(id<selectionIndicator[chosen_cutoff].Length()-1)
	       fprintf(VTout,"%g;",mafIncluded[id]);
	    else
	       fprintf(VTout,"%g",mafIncluded[id]);

	 }
      }
      fprintf(VTout,"\t");
      for(int id=0;id<selectionIndicator[chosen_cutoff].Length();id++)
      {
	 if(selectionIndicator[chosen_cutoff][id]==1)
	 {
	    if(id<selectionIndicator[chosen_cutoff].Length()-1)
	       fprintf(VTout,"%g;",singlevarPvalue[id]);
	    else
	       fprintf(VTout,"%g",singlevarPvalue[id]);
	 }
      }
      fprintf(VTout,"\t");
      for(int id=0;id<selectionIndicator[chosen_cutoff].Length();id++)
      {
	 if(selectionIndicator[chosen_cutoff][id]==1)
	 {
	    if(id<selectionIndicator[chosen_cutoff].Length()-1)
	       fprintf(VTout,"%g;",singlevarEff[id]);
	    else
	       fprintf(VTout,"%g",singlevarEff[id]);
	 }
      }
      fprintf(VTout,"\t%g",effSize[chosen_cutoff]);
      fprintf(VTout,"\t%g\n",pvalue);

      if(lower) delete [] lower;
      if(upper) delete [] upper;
      if(mean) delete [] mean;
   }

   fprintf(VTout,"===================================================================\n");
   fprintf(VTout,"\n%d %s tested.\n",group.annoGroups.Length(),group.annoGroups.Length()>1?"genes were":"gene was");
   fprintf(VTout,"There %s %d %s either had no rare allele in the qualified sample or did not pass the collapsing criterion.\n",bad>1?"are":"is", bad,bad>1?"genes":"gene");

   fclose(VTout);
}

double VT::CalculatePvalue(Eigen::VectorXf & score,Matrix & cov,double t_max)
{
   double pvalue=_NAN_;
   int dim = cov.cols;
   double chisq = t_max * t_max;

   StringDoubleHash jointProbHash;

   if(dim==1)
   {
      pvalue = pchisq(chisq,1,0,0);
      return pvalue;
   }
   double uni = pchisq(chisq, 1,0,0);
   pvalue += dim*uni;
   //printf("pvalue so far is:%g\n",pvalue);

   //Initializations for combination
   vector<int> indx;
   vector<int> alpha (dim,0);
   for(int i=0;i<dim;i++)
      alpha[i] = i;
   int n=dim;

   for(int r=2;r<=dim;r++)
   {
      int j=r;
      int k=r;

      Vector par,comb;
      par.Dimension(2);

      for(int twk=j;twk<=k;twk++)
      {
	 int r=twk;
	 bool done=true;
	 for(int iwk=0;iwk<r;iwk++)indx.push_back(iwk);
	 while(done)
	 {
	    done=false;
	    for(int owk=0;owk<r;owk++)
	    {
	       //A new combination has been generated!
	       comb.Push(alpha[indx[owk]]);
	    }
	    /*
	       printf("New combination:\n");
	       for(int i=0;i<comb.Length();i++)
	       printf("%g ",comb[i]);
	       printf("\n");
	     */
	    //Calculate P(A|BC)
	    par.Copy(GetConditionalDist(score,cov,comb));
	    //printf("pars are:%g %g\n",par[0],par[1]);
	    double chisq,condProb,prob;
	    if(par[1]==0.0)
	    {
	       condProb = 0.0;
	    }
	    else 
	    {
	       chisq = (t_max-par[0])*(t_max-par[0])/par[1];
	       if(chisq<0.0)
		  chisq = -chisq;
	       //condProb = chidist(chisq,1);
	       condProb = pchisq(chisq, 1,0,0);
	    }
	    //printf("conditional prob is:%g\n",condProb);
	    String hashKey;

	    if(r==2)
	    {
	       int tmp = comb[0];
	       hashKey += tmp;
	       tmp = comb[1];
	       hashKey += tmp;
	       //	       printf("hashKey is: %s\n",hashKey.c_str());
	       prob = condProb*uni;	       
	       //prob = exp(log(condProb) + log(uni));
	       //printf("joint prob is:%g\n",prob);
	       jointProbHash.SetDouble(hashKey,prob);
	       hashKey.Clear();
	    }
	    else
	    {
	       for(int i=1;i<r;i++)
	       {
		  int tmp = comb[i];
		  hashKey += tmp;
	       }
	       //	       printf("hashKey is: %s\n",hashKey.c_str());
	       prob = jointProbHash.Double(hashKey);
	       //printf("joint prob is:%g\n",prob);
	       //if(prob == _NAN_)
	       //{
	       //	  printf("Error! Joint probability %s is not hashed in properly.\n",hashKey);
	       //     }
	       prob *= condProb; //This is the new joint prob.
	       String newKey;
	       int tmp = comb[0];
	       newKey += tmp;
	       newKey += hashKey;
	       //	       printf("newKey is: %s\n",newKey.c_str());
	       jointProbHash.SetDouble(newKey,prob);
	       hashKey.Clear();
	    }
	    pvalue -= prob;
	    //printf("pvalue so far is:%g\n",pvalue);
	    comb.Clear();
	    for(int iwk=r-1;iwk>=0;iwk--)
	    {
	       if(indx[iwk]<=(n-1)-(r-iwk))
	       {
		  indx[iwk]++;
		  for(int swk=iwk+1;swk<r;swk++)
		  {
		     indx[swk]=indx[swk-1]+1;
		  }
		  iwk=-1;
		  done=true;
	       }
	    }
	 }
	 indx.clear();
      }
   }
   // printf("pvalue is:%g\n",pvalue);
   return pvalue;
}

//Get P(A|BC) where A is the first element in comb. Comb contains index only.
Vector VT::GetConditionalDist(Eigen::VectorXf & score,Matrix & cov,Vector & comb)
{
   Vector result;
   result.Dimension(2);
   result.Zero();

   //chop score vector and sigma matrix
   Vector mu2;
   Matrix sub_cov;
   int dim = comb.Length()-1;
   mu2.Dimension(dim);
   sub_cov.Dimension(dim,dim);
   for(int i=0;i<dim;i++)
   {
      int idx1 = comb[i+1];
      mu2[i] = score[idx1];
      for(int j=0;j<dim;j++)
      {
	 int idx2 = comb[j+1];
	 sub_cov[i][j] = cov[idx1][idx2];
      }
   }
   /*printf("sub_cov is:\n");
     for(int i=0;i<dim;i++)
     {
     for(int j=0;j<dim;j++)
     {
     printf("%g ",sub_cov[i][j]);
     }
     printf("\n");
     }*/

   Matrix inv;
   inv.Copy(sub_cov);
   SVD svd;
   svd.InvertInPlace(inv);
   Vector sigma12;
   sigma12.Dimension(dim);
   for(int i=0;i<dim;i++)
   {
      int idx1 = comb[0];
      int idx2 = comb[i+1];
      sigma12[i] = cov[idx1][idx2];
   }
   /*
      printf("sigma12 is:\n");
      for(int i=0;i<dim;i++)
      printf("%g ",sigma12[i]);
      printf("\n");
    */
   Vector tmp;
   tmp.Dimension(dim);
   tmp.Zero();
   for(int i=0;i<dim;i++)
   {
      tmp[i] += sigma12.InnerProduct(inv[i]);
   }

   //Calculate Mu and Sigma here
   result[0] = tmp.InnerProduct(mu2);
   result[1] = 1.0 - tmp.InnerProduct(sigma12); 

   if(result[1]<0)
      result[1] = fabs(result[1]);

   return result;
} 

void VT::Run(Pedigree & ped, FastTransform & trans,FastFit & engine,String & traitName,GroupFromAnnotation & group,FILE * log)
{
   if(!vcfAnnotated)
   {
      error("\nERROR! VCF file needs to be annotated in INFO field.\n");
   }

   if(asymptotic)
   {
      //Fill in UDX, UD, and UDY
      //trans.FinalizeTransform(engine.sigma_g2Hat, engine.sigma_e2Hat);

      if(!VT::VTwtMB && !VT::VTwtCMC_binary && !VT::VTwtCMC_counts)
      {
	 VTwtCMC_binary=true;
	 printf("Reminder: weight function was specified for Variable-threshold test. VTwtCMC_binary weight will be use in the following VT tests.\n");
      }

      fixedEff.resize(trans.persons);
      //      beta_hat.Dimension(trans.fixedEffects);
      //    engine.EstimateAnalyticalBeta(trans.UDY,trans, beta_hat);
      Eigen::VectorXf beta(engine.betaHat.Length());
      for(int i=0;i<beta.size();i++)
	 beta[i] = engine.betaHat[i];
      fixedEff = trans.UDX*beta;

      residuals_original = trans.UDY - fixedEff;
      sigma2_hat=1.0;
      //sigma2_hat = residuals_original.SumSquares()/trans.persons;

      if(SVT::dosage || SVT::genoFromVCF)
      {
	 reader.open(GroupFromAnnotation::vcfInput,header);
	 reader.readVcfIndex();
      }
      if(asymptotic)
      {
	 printf("\nPerforming VT tests ... ");
	 fflush(stdout);
	 CalculateAsymptotic(ped,trans,engine,traitName,group,log);
	 printf("completed.\n");
      }
   }
}

void VT::CalculateSingleVarPvalue(FastTransform & trans,String & geneName,FastFit & engine,Eigen::MatrixXf & genotypeSKAT)
{
   //printf("Working on singlevar on gene %s.\n",geneName.c_str());
   Eigen::MatrixXf transGeno_;
   transGeno_.resize(trans.persons,genotypeSKAT.cols());
   transGeno_ = trans.UD * genotypeSKAT;

   // Calculate singlevariant pvalues
   singlevarPvalue.Dimension(genotypeSKAT.cols(),0.0);
   singlevarEff.Dimension(genotypeSKAT.cols(),0.0);
   for(int m=0;m<genotypeSKAT.cols();m++)
   {
      Eigen::VectorXf geno;
      geno.resize(trans.persons);
      for(int p =0;p<trans.persons;p++)
      {
	 geno[p] = transGeno_(p,m);
      }
      double numerator = residuals_original.dot(geno);
      double denominator = geno.dot(geno);
      double pvalue =  pchisq(numerator*numerator/denominator,1,0,0);
      //double pvalue =  chidist(numerator*numerator/denominator,1);
      singlevarPvalue[m] = pvalue;
      singlevarEff[m] = numerator/denominator;
   }
   //printf("finished working on singlevar on gene %s.\n",geneName.c_str());
}
